#Standard python libraries
import os

#Dependencies
import numpy as np
import yaml
import matplotlib.pyplot as plt
import matplotlib
try:
    from pyfftw.interfaces.numpy_fft import fft, fftshift, ifft, ifftshift, fftfreq
except ImportError:
    from numpy.fft import fft, fftshift, ifft, ifftshift, fftfreq

def rescale_intensity(signal,factor):
    Smax = np.max(np.abs(signal))
    S = signal/Smax
    argS = np.angle(signal)
    logS = np.log10((np.abs(S)+1/factor))+np.log10(factor)
    logS = logS * Smax / np.max(np.abs(logS))
    return logS*np.exp(1j*argS)

def custom_colorbar(scale_function,cmap,cax,*,yticks=[]):
    n = 500
    x = np.arange(2) #dummy variable for x-axis
    y = np.linspace(-1,1,num=n)
    dy = y[1] - y[0]
    z = y
    Z = np.ones((y.size,x.size))*z[:,np.newaxis]
    
    y2 = np.linspace(-1-dy,1+dy,num=n+1) #Fix pcolormesh artifact
    
    cax.pcolormesh(x,y2,Z,cmap=cmap)
    cax.set_ylim([-1,1])
    cax.set_xticks([])
    if yticks:
        pass
    else:
        yticks = cax.get_yticks()
    cax.set_yticks(np.real(scale_function(yticks)))
    cax.set_yticklabels([str(x) for x in yticks])

def plot2D_logscaled(x,y,z,logfactor=1,part = 'real',norm = 1,vmax='max',ax=None,fig=None,
                     colorbar_ax=None,ticks=[],colormap='seismic',figsize=(4,3)):
    f = lambda x: rescale_intensity(x,logfactor)
    z = f(z)
    if fig == None and ax != None:
        raise Exception('Must pass both figure and axis handles, or neither')
    elif fig == None:
        fig = plt.figure(constrained_layout=True,figsize=figsize)
        widths = [1, 0.05]
        heights = [1]
        spec = fig.add_gridspec(ncols=len(widths), nrows=len(heights), width_ratios=widths,
                                height_ratios=heights)
        ax = fig.add_subplot(spec[0, 0])
        colorbar_ax = fig.add_subplot(spec[0,1])
    plot2D(x,y,z,part=part,norm=norm,fig=fig,ax=ax,colorbar=False,colormap=colormap,vmax=vmax)
    custom_colorbar(f,colormap,colorbar_ax,yticks=ticks)
    return fig, ax

def plot2D(x,y,z,contour_levels = None,part = 'complex',norm = 1,vmax='max',
           contours_only = False,contour_colormap='seismic',
           contour_colors=None,ax=None,fig=None,colorbar=True,
           colorbar_ax=None,ticks = [],alpha=None,colormap='seismic'):
    if ax == None:
        fig, ax = plt.subplots()
    elif fig == None:
        #ax was specified but not fig. Get the current figure
        fig = plt.gcf()
    X,Y = np.meshgrid(x,y,indexing='ij')
    if type(contour_levels) is np.ndarray:
        num_contours = contour_levels.size
        contour_flag = True
    elif type(contour_levels) is list:
        contour_levels = np.array(contour_levels)
        num_contours = contour_levels.size
        contour_flag = True
    else:
        contour_flag = False
        
    if part == 'real':
        Z = z.real
    elif part == 'imag':
        Z = z.imag
    elif part == 'abs':
        Z = np.abs(z)
    elif part == 'complex':
        Z = z
    if norm == 'self':
        norm = np.max(np.abs(Z))
    Z = Z/norm
    # print(np.max(np.abs(Z)))
    if contour_flag:
        if part == 'complex':
            contour_Z = np.abs(Z)
        else:
            contour_Z = Z
        if norm == 1:
            contour_scaling = np.max(np.abs(contour_Z))
        else:
            contour_scaling = 1
        contour_levels = contour_levels * contour_scaling
        if contour_colormap != None:
            handle = ax.contour(X,Y,contour_Z,num_contours,cmap=contour_colormap,
                                levels = contour_levels)
        else:
            handle = ax.contour(X,Y,contour_Z,num_contours,colors=contour_colors,
                                levels = contour_levels)
    if contours_only:
        if colorbar:
            set_colorbar(fig,handle,ticks,colorbar_ax)
    else:
        if part == 'complex':
            handle = ax.imshow(complex_array_to_rgb(Z.T,alpha=alpha), extent=(x[0],x[-1],y[0],y[-1]),origin='lower')
        else:
            if vmax == 'max':
                vmax = np.max(np.abs(Z))
            handle = ax.pcolormesh(X,Y,Z,cmap=colormap,vmin=-vmax,vmax=vmax,shading='Gouraud')
        if colorbar:
            set_colorbar(fig,handle,ticks,colorbar_ax)
    return fig, ax

def set_colorbar(fig,handle,ticks,cax):
    if ticks:
        if cax:
            fig.colorbar(handle,ticks=ticks,cax=cax)
        else:
            fig.colorbar(handle,ticks=ticks)
    else:
        if cax:
            fig.colorbar(handle,cax=cax)
        else:
            fig.colorbar(handle)

def complex_array_to_rgb(X, theme='light', rmax=None,alpha = None):
    '''Taken from https://stackoverflow.com/questions/15207255/is-there-any-way-to-use-bivariate-colormaps-in-matplotlib/17113417#17113417
    Takes an array of complex number and converts it to an array of [r, g, b],
    where phase gives hue and saturaton/value are given by the absolute value.
    Especially for use with imshow for complex plots.'''
    absmax = rmax or np.abs(X).max()
    Y = np.zeros(X.shape + (3,), dtype='float')
    Y[..., 0] = np.angle(X) / (2 * np.pi) % 1
    if theme == 'light':
        Y[..., 1] = np.clip(np.abs(X) / absmax, 0, 1)
        Y[..., 2] = 1
    elif theme == 'dark':
        Y[..., 1] = 1
        Y[..., 2] = np.clip(np.abs(X) / absmax, 0, 1)
    Y = matplotlib.colors.hsv_to_rgb(Y)
    if alpha == None:
        print(alpha)
        pass
    else:
        newY = np.zeros(X.shape + (4,), dtype=Y.dtype)
        newY[...,:3] = Y
        newY[...,3] = 1
        for i in range(Y.shape[0]):
            for j in range(Y.shape[1]):
                if np.all(Y[i,j,:3] == 1):
                    newY[i,j,3] = alpha
        Y = newY
    return Y

class Plotter(object):
    def universal_plotting_commands(self,create_figure=True,save_name=None):
        if create_figure:
            plt.figure()
        if save_name != None:
            plt.savefig(save_name)

    def plotTA(self,*,frequency_range=[-1000,1000], subtract_DC = True, create_figure=True,
               color_range = 'auto',draw_colorbar = True,save_fig=True,scale_factor = 1,
               delay_time_start = -100,save_name='default'):
        """Plots the transient absorption spectra with detection frequency on the
        y-axis and delay time on the x-axis.

        Args:
            frequency_range (list): sets the min (list[0]) and max (list[1]) detection frequency for y-axis
            subtract_DC (bool): if True subtract the DC component of the TA
            color_range (list): sets the min (list[0]) and max (list[1]) value for the colorbar
            draw_colorbar (bool): if True add a colorbar to the plot
            save_fig (bool): if True save the figure that is produced
        """
        # Cut out unwanted detection frequency points
        w_ind = np.where((self.w > frequency_range[0]) & (self.w < frequency_range[1]))[0]
        w = self.w[w_ind]
        sig = self.signal[:,w_ind]*scale_factor

        if subtract_DC:
            T, sig = self.remove_DC(self.delay_times,sig,axis=0)
            tt, ww = np.meshgrid(T,w,indexing='ij')
        else:
            tt, ww = np.meshgrid(self.delay_times,w,indexing='ij')
        if create_figure:
            plt.figure()
        if color_range == 'auto':
            plt.pcolormesh(tt,ww,np.real(sig))
        else:
            plt.pcolormesh(tt,ww,np.real(sig),vmin=color_range[0],vmax=color_range[1])
        if draw_colorbar:
            plt.colorbar()
        plt.xlabel('Delay time ($\omega_0^{-1}$)',fontsize=16)
        plt.ylabel('Detection Frequency ($\omega_0$)',fontsize=16)
        if save_fig:
            if save_name == 'default':
                plt.savefig(os.path.join(self.base_path,'TA_spectra_iso_ave'))
            else:
                plt.savefig(os.path.join(self.base_path,save_name))
            
    def plotTA_ft(self,*,delay_time_start = 1,create_figure=True,color_range = 'auto',subtract_DC=True,
                   draw_colorbar = True,frequency_range=[-1000,1000],normalize=False,part='abs',
                   save_fig=True,wT_frequency_range = 'auto',omega0 = 1):
        w_ind = np.where((self.w > frequency_range[0]) & (self.w < frequency_range[1]))[0]
        w = self.w[w_ind] * omega0
        sig = self.signal[w_ind,:]

        delay_time_indices = np.where(self.delay_times > delay_time_start)[0]
        delay_times = self.delay_times[delay_time_indices]
        sig = sig[:,delay_time_indices]
        if normalize:
            sig /= np.dot(self.dipoles,self.dipoles)**2
        wT = fftshift(fftfreq(delay_times.size,d=(delay_times[1] - delay_times[0])))*2*np.pi*omega0
        sig_fft = fft(sig,axis=1)
        if subtract_DC:
            sig_fft[:,0] = 0
        sig_fft = fftshift(sig_fft,axes=(1))
        
        ww, wTwT = np.meshgrid(wT,w)

        if create_figure:
            plt.figure()

        if part == 'real':
            plt.title('Real Part')
            plot_sig = np.real(sig_fft)
        elif part == 'imag':
            plt.title('Imag Part')
            plot_sig = np.imag(sig_fft)
        elif part == 'abs':
            plt.title('Magnitude')
            plot_sig = np.abs(sig_fft)
        elif part == 'phase':
            plt.title('Phase')
            plot_sig = np.arctan2(np.imag(sig_fft),np.real(sig_fft))
        else:
            raise Exception('Unknown part keyword argument')
        if color_range == 'auto':
            plt.pcolormesh(ww,wTwT,plot_sig)
        else:
            plt.pcolormesh(ww,wTwT,plot_sig,vmin=color_range[0],vmax=color_range[1])
        if draw_colorbar:
            plt.colorbar()
        if omega0 == 1:
            plt.xlabel('$\omega_T$ ($\omega_0$)',fontsize=16)
            plt.ylabel('Detection Frequency ($\omega_0$)',fontsize=16)
        else:
            plt.xlabel('$\omega_T$ ($cm^{-1}$)',fontsize=16)
            plt.ylabel('Detection Frequency ($cm^{-1}$)',fontsize=16)
            
        if wT_frequency_range == 'auto':
            plt.xlim([0,np.max(wT)])
        else:
            plt.xlim(wT_frequency_range)
        if save_fig:
            plt.savefig(self.base_path + 'TA_spectra_fft')

    def load_eigen_params(self):
        with open(os.path.join(self.base_path,'eigen_params.yaml'),'r') as yamlstream:
            eigen_params = yaml.load(yamlstream,Loader=yaml.SafeLoader)
            self.truncation_size = eigen_params['final truncation size']
            self.ground_ZPE = eigen_params['ground zero point energy']
            self.ground_to_excited_transition = eigen_params['ground to excited transition']

    def plotTA_units(self,*,frequency_range=[-1000,1000], subtract_DC = True, create_figure=True,
                           color_range = 'auto',draw_colorbar = True,save_fig=True,omega_0 = 1):
        """Plots the transient absorption spectra with detection frequency on the
        y-axis and delay time on the x-axis.

        Args:
            frequency_range (list): sets the min (list[0]) and max (list[1]) detection frequency for y-axis
            subtract_DC (bool): if True subtract the DC component of the TA
            color_range (list): sets the min (list[0]) and max (list[1]) value for the colorbar
            draw_colorbar (bool): if True add a colorbar to the plot
            save_fig (bool): if True save the figure that is produced
            omega_0 (float): convert from unitless variables, omega_0 should be provided in wavenumbers
        """
        self.load_eigen_params()
        f0_thz = omega_0 * 3E10/1.0E12 # omega_0 in wavenumbers
        T_ps = self.delay_times / f0_thz / (2*np.pi)
        # Cut out unwanted detection frequency points
        self.w += self.ground_to_excited_transition + self.center
        self.w *= omega_0
        w_ind = np.where((self.w > frequency_range[0]) & (self.w < frequency_range[1]))[0]
        w = self.w[w_ind]
        sig = self.signal[w_ind,:]

        if omega_0 == 1:
            xlab = r'Delay time ($\omega_0^{-1}$)'
            ylab = r'Detection Frequency ($\omega_0$)'
        else:
            xlab = 'Delay time (ps)'
            ylab = r'Detection Frequency (cm$^{-1}$)'

        if subtract_DC:
            sig_fft = fft(sig,axis=1)
            sig_fft[:,0] = 0
            sig = np.real(ifft(sig_fft))
            ww, tt = np.meshgrid(T_ps,w)
        if create_figure:
            plt.figure()
        if color_range == 'auto':
            plt.pcolormesh(ww,tt,sig)
        else:
            plt.pcolormesh(ww,tt,sig,vmin=color_range[0],vmax=color_range[1])
        if draw_colorbar:
            plt.colorbar()
        plt.xlabel(xlab,fontsize=16)
        plt.ylabel(ylab,fontsize=16)
        if save_fig:
            plt.savefig(self.base_path + 'TA_spectra_iso_ave')

    def plot_integrated_TA_units(self,*, create_figure=True,save_fig=True,omega_0 = 1,marker='-o'):
        """Plots the transient absorption spectra with detection frequency on the
        y-axis and delay time on the x-axis.

        Args:
            frequency_range (list): sets the min (list[0]) and max (list[1]) detection frequency for y-axis
            subtract_DC (bool): if True subtract the DC component of the TA
            color_range (list): sets the min (list[0]) and max (list[1]) value for the colorbar
            draw_colorbar (bool): if True add a colorbar to the plot
            save_fig (bool): if True save the figure that is produced
            omega_0 (float): convert from unitless variables, omega_0 should be provided in wavenumbers
        """
        self.load_eigen_params()
        f0_thz = omega_0 * 3E10/1.0E12 # omega_0 in wavenumbers
        T_ps = self.delay_times / f0_thz / (2*np.pi)
        
        sig = np.trapz(self.signal,x=self.w,axis=0)


        xlab = 'Delay time (ps)'

        if create_figure:
            plt.figure()
        plt.plot(T_ps,sig,marker)
        plt.xlabel('Delay time (ps)',fontsize=16)
        plt.ylabel('$S_{TA,FI}(T)$ (arb.)',fontsize=16)
        if save_fig:
            plt.savefig(self.base_path + 'TA_spectra_iso_ave')

    def plot_integrated_signal(self,create_figure = True,scale_factor = 1,
                               frequency_range = 'all',save_fig = False,save_name='default'):
        if create_figure:
            plt.figure()
        plt.plot(self.delay_times,self.integrated_signal(frequency_range=frequency_range)*scale_factor)
        plt.xlabel('Delay time ($\omega_0^{-1}$)')
        plt.ylabel('Frequency-Integrated TA')
        if save_fig:
            if save_name == 'default':
                plt.savefig(os.path.join(self.base_path,'TA_integrated_iso_ave'))
            else:
                plt.savefig(os.path.join(self.base_path,save_name))
            

    def plot_T_slice(self,T,*,create_figure = True,scale_factor = 1,subtract_DC = True):
        if subtract_DC:
            sig = self.subtract_DC(self.signal)
        else:
            sig = self.signal
        T_ind, T = self.get_closest_index_and_value(T,self.delay_times)
        sig = sig[:,T_ind] * scale_factor
        if create_figure:
            plt.figure()
        plt.title('T = {:.2f}'.format(T))
        plt.plot(self.w,sig)

    def plot_w_slice(self,w,*,create_figure = True,scale_factor = 1,subtract_DC = True):
        if subtract_DC:
            sig = self.subtract_DC(self.signal)
        w_ind, w = self.get_closest_index_and_value(w,self.w)
        sig = sig[w_ind,:] * scale_factor
        if create_figure:
            plt.figure()
        plt.title('w = {:.2f}'.format(w))
        plt.plot(self.delay_times,sig)

    def plot_wT_slice(self,wT,*,create_figure = True,scale_factor = 1,subtract_DC = True,part='real'):
        sig = self.ft_axis1(self.signal,zero_DC = subtract_DC)
        wT_ind,wT = self.get_closest_index_and_value(wT,self.wT)
        sig = sig[:,wT_ind]
        if part == 'real':
            plt_title = 'Real Part'
            plot_sig = np.real(sig)
        elif part == 'imag':
            plt_title = 'Imag Part'
            plot_sig = np.imag(sig)
        elif part == 'abs':
            plt_title = 'Magnitude'
            plot_sig = np.abs(sig)
        elif part == 'phase':
            plt_title = 'Phase'
            plot_sig = np.arctan2(np.imag(sig),np.real(sig))
        else:
            raise Exception('Unknown part keyword argument')
        if create_figure:
            plt.figure()
        plt_title += ', $\omega_T$ = {:.2f}'.format(wT)
        plt.title(plt_title)
        plt.plot(self.w,plot_sig,'.')



#Standard python libraries
import os
import warnings
import copy
import time
import itertools

#Dependencies - numpy, scipy, matplotlib, pyfftw
import numpy as np
import matplotlib.pyplot as plt
import pyfftw
from pyfftw.interfaces.numpy_fft import fft, fftshift, ifft, ifftshift, fftfreq
import scipy
from scipy.interpolate import interp1d as sinterp1d


class HeavisideConvolve:
    """This class calculates the discrete convolution of an array with the 
        heaviside step function

    Attributes:
        size (int) : number of linear convolution points
        theta_fft (numpy.ndarray) : discrete fourier transform of the step 
            function
        a : aligned array of zeros for use with the fftw algorithm
        b : empty aligned array for use with the fftw algorithm
        c : empty aligned array for use with the fftw algorithm
        fft : method for calculating the FFT of a (stores the result in b)
        ifft : method for calculating the IFFT of b (stores the result in c)
        
"""
    def __init__(self,arr_size):
        """
        Args:
            arr_size (int) : number of points desired for the linear 
                convolution
"""
        self.size = arr_size
        self.theta_fft = self.heaviside_fft()
        # The discrete convolution is inherently circular. Therefore we
        # perform the convolution using 2N-1 points
        self.a = pyfftw.empty_aligned(2*self.size - 1, dtype='complex128', n=16)
        self.b = pyfftw.empty_aligned(2*self.size - 1, dtype='complex128', n=16)
        self.c = pyfftw.empty_aligned(2*self.size - 1, dtype='complex128', n=16)
        pyfftw.interfaces.cache.enable()
        pyfftw.interfaces.cache.set_keepalive_time(10)
        self.fft = pyfftw.FFTW(self.a, self.b)
        self.ifft = pyfftw.FFTW(self.b,self.c,direction='FFTW_BACKWARD')
        self.a[:] = 0

    def heaviside_fft(self,*,value_at_zero=0.5):
        """This method calculates the FFT of the heaviside step function
        
        Args:
            value_at_zero (float): value of the heaviside step function at 
                x = 0

        Returns:
            numpy.ndarray: the FFT of the heaviside step function
"""
        # The discrete convolution is inherently circular. Therefore we perform the
        # convolution using 2N-1 points. Spacing dx is irrelevant for evaluating
        # the heaviside step function. However it is crucial that the x = 0 is included
        t = np.arange(-self.size+1,self.size)
        y = np.heaviside(t,value_at_zero)
        return fft(y)

    def fft_convolve(self,arr,*,d=1):
        """This method calculates the linear convolution of an input with 
            the heaviside step function
        
        Args:
            arr (numpy.ndarray): 1d array of input function values f(x)
            d (float): spacing size of grid f(x) is evaluated on, dx

        Returns:
            numpy.ndarray: linear convolution of arr with heaviside step 
                function
"""
        self.a[:arr.size] = arr

        self.b = self.fft()
        
        self.b *= self.theta_fft

        self.b = self.b

        self.c = self.ifft()
        # Return only the results of the linear convolution
        return self.c[-arr.size:] * d

    def fft_convolve2(self,arr,*,d=1):
        """This method loops over fft_convolve in order to perform the convolution of input array with the heaviside step function along the second axis of arr

        Args:
            arr (numpy.ndarray): 2d array of input function values f_i(x), 
                where i is the 1st index of the array
            d (float): spacing size of grid f_i(x) is evaluated on, dx

        Returns:
            numpy.ndarray: 2d array of linear convolution of arr with 
                heaviside step function along the second axis of arr
"""
        size0,size1 = arr.shape
        for i in range(size0):
            arr[i,:] = self.fft_convolve(arr[i,:],d=d)
        return arr

class HeavisideConvolveSP:
    """This class calculates the discrete convolution of an array with the 
        heaviside step function times an exponential decay/oscillation 
        envelope

    Attributes:
        size (int) : number of linear convolution points
        theta_fft (numpy.ndarray) : discrete fourier transform of the step 
            function
        a : aligned array of zeros for use with the fftw algorithm
        b : empty aligned array for use with the fftw algorithm
        c : empty aligned array for use with the fftw algorithm
        fft : method for calculating the FFT of a (stores the result in b)
        ifft : method for calculating the IFFT of b (stores the result in c)
        
"""
    def __init__(self,arr_size):
        """
        Args:
            arr_size (int) : number of points desired for the linear 
                convolution
"""
        self.size = arr_size
        # The discrete convolution is inherently circular. Therefore we
        # perform the convolution using 2N-1 points
        self.a = pyfftw.empty_aligned(2*self.size - 1, dtype='complex128', n=16)
        self.b = pyfftw.empty_aligned(2*self.size - 1, dtype='complex128', n=16)
        self.c = pyfftw.empty_aligned(2*self.size - 1, dtype='complex128', n=16)

        self.a_theta = pyfftw.empty_aligned(2*self.size - 1, dtype='complex128', n=16)
        self.b_theta = pyfftw.empty_aligned(2*self.size - 1, dtype='complex128', n=16)
        
        pyfftw.interfaces.cache.enable()
        pyfftw.interfaces.cache.set_keepalive_time(10)
        self.fft = pyfftw.FFTW(self.a, self.b)
        self.ifft = pyfftw.FFTW(self.b,self.c,direction='FFTW_BACKWARD')
        self.a[:] = 0

        t = np.arange(-self.size+1,self.size)
        self.theta = np.heaviside(t,0.5)
    
        self.fft_theta = pyfftw.FFTW(self.a_theta, self.b_theta)
        self.a_theta[:] = 0

    def fft_convolve(self,arr1,arr2,*,d=1):
        """This method calculates the linear convolution of an input with 
            the heaviside step function
        
        Args:
            arr (numpy.ndarray): 1d array of input function values f(x)
            d (float): spacing size of grid f(x) is evaluated on, dx

        Returns:
            numpy.ndarray: linear convolution of arr with heaviside step 
                function
"""
        self.a[:arr2.size] = arr2
        self.b = self.fft()

        self.a_theta[:] = self.theta
        self.a_theta[-arr1.size:] *= arr1

        self.b_theta = self.fft_theta()
        
        self.b *= self.b_theta

        self.b = self.b

        self.c = self.ifft()
        # Return only the results of the linear convolution
        return self.c[-arr2.size:] * d

    def fft_convolve2(self,arr1,arr2,*,d=1):
        """This method loops over fft_convolve in order to perform the convolution of input array with the heaviside step function along the second axis of arr

        Args:
            arr (numpy.ndarray): 2d array of input function values f_i(x), 
                where i is the 1st index of the array
            d (float): spacing size of grid f_i(x) is evaluated on, dx

        Returns:
            numpy.ndarray: 2d array of linear convolution of arr with 
                heaviside step function along the second axis of arr
"""
        size0,size1 = arr2.shape
        for i in range(size0):
            arr2[i,:] = self.fft_convolve(arr1[i,:],arr2[i,:],d=d)
        return arr2



import os
import time
import yaml

import numpy as np

from ufss import CompositeDiagrams

class ClosedBaseClass(CompositeDiagrams):
    def __init__(self,*,detection_type='polarization'):
        CompositeDiagrams.__init__(self,detection_type=detection_type)
        
        self.K_dict = {'u':self.up,'d':self.down}

        self.psis = dict()
        self.composite_psis = dict()

    def set_min_max_manifolds(self):
        if 'all_manifolds' in self.manifolds:
            pass
        else:
            manifold_integers = [self.manifold_key_to_number(man) for man
                                 in self.manifolds]
            self.maximum_manifold = max(manifold_integers)
            self.minimum_manifold = min(manifold_integers)

    def up(self,psi_in,*,next_manifold_mask = None,pulse_number = 0):
        """This method connects psi_p to psi_pj where the next order psi
            is one manifold above the current manifold.

        Args:
            psi_in (psi_container): input density matrix
            pulse_number (int): index of optical pulse (0,1,2,...) can also be set to
                'impulsive'
            next_manifold_mask (np.ndarray): optional - define the states to be considered 
                in the next manifold

        Returns:
            (psi_container): output from method next_order
"""

        return self.next_order(psi_in,up_flag=True,
                               next_manifold_mask = next_manifold_mask,
                               pulse_number = pulse_number)

    def down(self,psi_in,*,next_manifold_mask = None,pulse_number = 0):
        """This method connects psi_p to psi_pj where the next order psi
            is one manifold above the current manifold.

        Args:
            psi_in (psi_container): input density matrix
            pulse_number (int): index of optical pulse (0,1,2,...) can also be set to
                'impulsive'
            next_manifold_mask (np.ndarray): optional - define the states to be considered 
                in the next manifold

        Returns:
            (psi_container): output from method next_order
"""

        return self.next_order(psi_in,up_flag=False,
                               next_manifold_mask = next_manifold_mask,
                               pulse_number = pulse_number)

    def execute_diagram(self,instructions):
        num_instructions = len(instructions['ket'])+len(instructions['bra'])
        
        ket = self.psi0
        bra = self.psi0
        ketname = ''
        braname = ''
        ket_instructions = instructions['ket']
        bra_instructions = instructions['bra']
        for i in range(len(ket_instructions)):
            key, num = ket_instructions[i]
            ketname += key+str(num)
            # Try to re-use previous calculations, if they exist
            try:
                new_ket = self.psis[ketname]
            except KeyError:
                new_ket = self.K_dict[key](ket,pulse_number=num)
                self.psis[ketname] = new_ket
            ket = new_ket
                

        for i in range(len(bra_instructions)):
            key, num = bra_instructions[i]
            braname += key+str(num)
            # Try to re-use previous calculations, if they exist
            try:
                new_bra = self.psis[braname]
            except KeyError:
                new_bra = self.K_dict[key](bra,pulse_number=num)
                self.psis[braname] = new_bra
            bra = new_bra

        rho = (bra,ket)
        return rho

    def check_pulse_ordering_and_overlap(self,pdc_in,pdc_ref,pulse_num):
        """Have to replace the original method here, since some of the logic
            in the original method only applies for density matrix 
            calculations
"""
        remaining_pdc = pdc_ref - pdc_in
        if np.any(remaining_pdc < 0):
            # input pdc has more interactions with a pulse than reference
            return False
        else:
            return True

    def test_RWA(self,old_manifold_key,ins):
        if 'all_manifolds' in self.manifolds:
            test = True
        else:
            man_num = self.manifold_key_to_number(old_manifold_key)
            if ins[0][-1] == 'u':
                change = 1
            else:
                change = -1
            new_man_num = man_num + change
            if new_man_num < self.minimum_manifold:
                test = False
            elif new_man_num > self.maximum_manifold:
                test = False
            else:
                test = True

        return test

    def execute_composite_diagrams(self):
        self.composite_psis[self.psi0.pdc_tuple] = self.psi0
        psis = self.composite_psis
        old_psis = [self.psi0]
            
        for i in range(self.highest_order):
            new_psis = []
            pdcs = set()
            for old_psi in old_psis:
                old_pdc = old_psi.pdc_tuple
                next_instructions = self.get_next_interactions(old_pdc)
                for key in next_instructions.keys():
                    ket_ins, bra_ins = next_instructions[key]
                    test = self.test_RWA(old_psi.manifold_key,ket_ins)
                    if not test:
                        pass
                    else:
                        pdcs.add(key)
                        if key in psis:
                            # do not do calculations if they are already done
                            pass
                        else:
                            new_psi=self.execute_interaction(old_psi,ket_ins)
                            new_psis.append(new_psi)
            
            for new_psi in new_psis:
                if new_psi.pdc_tuple in psis.keys():
                    partial_psi = psis[new_psi.pdc_tuple]
                    new_psi = self.add_psis(partial_psi,new_psi)
                psis[new_psi.pdc_tuple] = new_psi

            old_psis = [psis[pdc] for pdc in pdcs]

        return None

    def execute_interaction(self,psi_in,interaction):
        """This method connects the input density matrix of order n to a
            density matrix that contributes to order n + 1, according to the
            specified interaction type and pulse number
        Args:
            rho_in (DensityMatrices) : input density matrix
            interaction (tuple) : specifies interaction type and pulse 
                number in format (str, int)
"""
        int_type, pulse_num = interaction
        int_type = int_type.strip('K')
        return self.K_dict[int_type](psi_in,pulse_number = pulse_num)

    def set_signal_pdcs(self,pdcs):
        """Specific to closed systems, needed for calculating wavefunctions
        Args:
            pdcs (list) : list of pdcs as np.ndarrays or tuples
"""
        if type(pdcs[0]) is tuple:
            self.singal_pdcs = pdcs
            signal_arrs = [self.pdc_tup_to_arr(pdc) for pdc in pdcs]
        else:
            signal_arrs = pdcs
            self.signal_pdcs = [self.pdc_arr_to_tup(pdc) for pdc in pdcs]
        
        cc_signal_pdcs = [self.pdc_arr_to_tup(arr[:,::-1])
                          for arr in signal_arrs]
        self.all_pdcs = list(set(self.signal_pdcs + cc_signal_pdcs))

        sh = signal_arrs[0].shape
        sh = sh + (len(self.all_pdcs),)
        all_pdcs_arr = np.zeros(sh,dtype='int')
        for i in range(len(self.all_pdcs)):
            all_pdcs_arr[...,i] = self.pdc_tup_to_arr(self.all_pdcs[i])
        self.max_pdc = np.max(all_pdcs_arr,axis=-1)
        self.set_interaction_types()

        return None

    def remove_calculations_by_pulse_number(self,pulse_number):
        num = str(pulse_number)
        keys = self.psis.keys()
        keys_to_remove = []
        for key in keys:
            flag = key.find(num)
            if flag >= 0:
                keys_to_remove.append(key)
        for key in keys_to_remove:
            self.psis.pop(key)

    def remove_composite_calculations_by_pulse_number(self,pulse_number):
        keys = self.composite_psis.keys()
        keys_to_remove = []
        for key in keys:
            r_ints,cr_ints = key[pulse_number]
            if r_ints > 0 or cr_ints > 0:
                keys_to_remove.append(key)
        for key in keys_to_remove:
            self.composite_psis.pop(key)

    def set_current_diagram_instructions(self,arrival_times):
        self.current_instructions = self.get_wavefunction_diagram_dict(arrival_times)

    # def fluorescence_detection_signal(self,bra_dict,ket_dict,*,time_index = -1):
    #     """Calculate inner product given an input bra and ket dictionary
    #         at a time given by the time index argument.  Default is -1, since
    #         2DFS is concerned with the amplitude of arriving in the given manifold
    #         after the pulse has finished interacting with the system."""
    #     bra = np.conjugate(bra_dict['psi'][:,-1])
    #     ket = ket_dict['psi'][:,-1]
    #     return np.dot(bra,ket)

    def reset(self):
        self.psis = dict()
        self.composite_psis = dict()

    def manifold_key_to_number(self,key):
        num = self.ordered_manifolds.index(key)
        return num

    def manifold_number_to_key(self,num):
        if num < 0:
            raise Exception('Manifold number must be positive')
        key = self.ordered_manifolds[num]
        return key

    def get_new_manifold_key(self,old_manifold_key,up_flag):
        if up_flag:
            change = 1
        else:
            change = -1
        if 'all_manifolds' in self.manifolds:
            new_manifold_key = 'all_manifolds'
        else:
            old_manifold = self.manifold_key_to_number(old_manifold_key)
            new_manifold = old_manifold + change
            new_manifold_key = self.manifold_number_to_key(new_manifold)
        return new_manifold_key
    
    def get_H_mu(self,pulse_number,key,rotating_flag=True):
        """Calculates the dipole matrix given the electric field 
            polarization vector"""
        t0 = time.time()
        pol = self.polarization_sequence[pulse_number]

        x = np.array([1,0,0])
        y = np.array([0,1,0])
        z = np.array([0,0,1])
        try:
            mu = self.H_mu[key]
        except KeyError:
            try:
                key = 'up'
                mu = self.H_mu[key]
            except KeyError:
                key = 'ket_up'
                mu = self.H_mu[key]
        if type(mu) is list:
            if np.all(pol == x):
                overlap_matrix = mu[0]#.copy()
            elif np.all(pol == y):
                overlap_matrix = mu[1]#.copy()
            elif np.all(pol == z):
                overlap_matrix = mu[2]#.copy()
            else:
                overlap_matrix = mu[0]*pol[0] + mu[1]*pol[1] + mu[2]*pol[2]
        else:
            if np.all(pol == x):
                overlap_matrix = mu[:,:,0]#.copy()
            elif np.all(pol == y):
                overlap_matrix = mu[:,:,1]#.copy()
            elif np.all(pol == z):
                overlap_matrix = mu[:,:,2]#.copy()
            else:
                overlap_matrix = np.tensordot(mu,pol,axes=(-1,0))

        if not rotating_flag:
            overlap_matrix = overlap_matrix.T

        t1 = time.time()
        self.dipole_time += t1-t0

        return overlap_matrix



import numpy as np
import ufss
import os
import yaml
import unittest

def direction_cosine_matrix(phi,th,chi):
    c = np.cos
    s = np.sin
    M = np.array([
        [c(phi)*c(th)*c(chi)-s(phi)*s(chi), s(phi)*c(th)*c(chi)+c(phi)*s(chi), -s(th)*c(chi)],
        [-c(phi)*c(th)*s(chi)-s(phi)*c(chi), -s(phi)*c(th)*s(chi)+c(phi)*c(chi), s(th)*s(chi)],
        [c(phi)*s(th), s(phi)*s(th), c(th)]
        ])
    return M

def manual_average(sc,lab_polarizations,num_angles):
    sc.reset()
    sc.polarization_sequence = lab_polarizations
    signal = sc.calculate_signal_all_delays()
    signal[...] = 0
    dtheta = 2*np.pi/num_angles
    for i in range(0,num_angles//2):
        theta = dtheta*i
        for j in range(0,num_angles):
            phi = dtheta*j
            for k in range(0,num_angles):
                chi = dtheta * k
                dcm = direction_cosine_matrix(phi,theta,chi)

                new_polarizations = [dcm.dot(pol) for pol in lab_polarizations]
                sc.polarization_sequence = new_polarizations
                sc.reset()
                signal += sc.calculate_signal_all_delays()*np.sin(theta)*dtheta**3

    return signal/(8*np.pi**2)

def run_HLG():
    site_energies = [100,100.5]#*5
    site_couplings = [0.5]#[0.5,0,0.2,0.1,0]*9
    dipoles = [[1,0,0],
            [0,1,0]]#*5

    trunc_size = 1

    folder = 'fixtures/OrthogonalDimer'
    os.makedirs(folder,exist_ok=True)

    site_bath = {'cutoff_frequency':1,
                'coupling':0.05,
                'temperature':0,
                'cutoff_function':'lorentz-drude',
                    'spectrum_type':'ohmic'}

    vibration_bath = site_bath # use same bath for sites and vibrations

    relax_bath = {'dephasing_rate':0.05, # additonal "pure" dephasing
                'relaxation_rate':0.05,
                'temperature':0,
                    'spectrum_type':'white-noise'}


    Redfield_bath = {'secular':True,
            'site_bath':site_bath,
            'vibration_bath':vibration_bath}

    #include if you wish to model inter-manifold relaxation
    Redfield_bath['site_internal_conversion_bath'] = relax_bath

    vibrations = []

    params = {
        'site_energies':site_energies,
        'site_couplings':site_couplings,
        'dipoles':dipoles,
        'truncation_size':trunc_size,
        'vibrations':vibrations}

    params['bath'] = Redfield_bath

    with open(os.path.join(folder,'simple_params.yaml'),'w+') as new_file:
        yaml.dump(params,new_file)

    ufss.HLG.run(folder,conserve_memory=False)
    return folder

class test_iso_ave(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        folder = run_HLG()

        TA4 = ufss.signals.UF2Wavefunctions(os.path.join(folder,'closed'))
        TA4_pdc = ((1,0),(0,1),(1,0))
        TA4_times = [np.array([0])]*4
        TA4_efields = [np.array([1])]*4
        TA4_centers = [100]*4
        TA4.set_efields(TA4_times,TA4_efields,TA4_centers,TA4_pdc)

        TA4.gamma_res = 10
        TA4.set_t(0.1)

        T = np.arange(1,10)

        TA4.set_pulse_delays([np.array([0]),T])

        TA4.polarization_sequence= [np.array([1,0,0])]*4

        cls.TA4 = TA4

        cls.TA4_iso = ufss.signals.IsotropicAverage(cls.TA4)
        # running the below calculates all the necessary molecular frame 
        # polarizations and saves them, so that they don't have to be
        # redone for each test below
        cls.TA4_iso.averaged_signal(['x','x','x','x'],return_signal=True)[0,...]

        cls.rel_tol = 0.01

    def test_xx(self):
        sig_averaged_xx = self.TA4_iso.averaged_signal(['x','x','x','x'],
                                                       return_signal=True)[0,...]
        xxxx = [np.array([0,0,1])]*4
        sig_xx_manual_averaging = manual_average(self.TA4,xxxx,8)[0,:,:]
        abs_diff = np.max(np.abs(sig_xx_manual_averaging-sig_averaged_xx))
        rel_diff = abs_diff/np.max(np.abs(sig_averaged_xx))
        self.assertTrue(rel_diff<self.rel_tol)

    def test_zy(self):
        sig_averaged_xy = self.TA4_iso.averaged_signal(['x','x','y','y'],
                                                       return_signal=True)[0,...]
        #zzyy converges much faster than xxyy, but converges to the same result
        zzyy = [np.array([0,0,1])]*2 + [np.array([0,1,0])]*2
        sig_zy_manual_averaging = manual_average(self.TA4,zzyy,10)[0,:,:]
        abs_diff = np.max(np.abs(sig_zy_manual_averaging-sig_averaged_xy))
        rel_diff = abs_diff/np.max(np.abs(sig_averaged_xy))
        self.assertTrue(rel_diff<self.rel_tol)
    

class test_iso_ave_5th_order(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        folder = run_HLG()

        TA6 = ufss.signals.UF2Wavefunctions(os.path.join(folder,'closed'))
        TA6_pdc = ((1,0),(1,0),(0,1),(0,1),(1,0))
        TA6_times = [np.array([0])]*6
        TA6_efields = [np.array([1])]*6
        TA6_centers = [100]*6
        TA6.set_efields(TA6_times,TA6_efields,TA6_centers,TA6_pdc)

        TA6.gamma_res = 10
        TA6.set_t(0.1)

        T = np.arange(1,10)

        TA6.set_pulse_delays([np.array([0]),np.array([0]),np.array([0]),T])

        TA6.polarization_sequence= [np.array([1,0,0])]*4

        cls.TA6 = TA6

        cls.TA6_iso = ufss.signals.IsotropicAverage(cls.TA6)
        # running the below calculates all the necessary molecular frame 
        # polarizations and saves them, so that they don't have to be
        # redone for each test below
        cls.TA6_iso.averaged_signal(['x','x','x','x','x','x'],return_signal=True)[0,0,0,...]

        cls.rel_tol = 0.01

    def test_zz(self):
        sig_averaged_xx = self.TA6_iso.averaged_signal(['x','x','x','x','x','x'],
                                                       return_signal=True)[0,0,0,...]
        z6 = [np.array([0,0,1])]*6
        sig_xx_manual_averaging = manual_average(self.TA6,z6,8)[0,0,0,:,:]
        abs_diff = np.max(np.abs(sig_xx_manual_averaging-sig_averaged_xx))
        rel_diff = abs_diff/np.max(np.abs(sig_averaged_xx))
        print(rel_diff)
        self.assertTrue(rel_diff<self.rel_tol)

    def test_zy(self):
        sig_averaged_xy = self.TA6_iso.averaged_signal(['x','x','x','x','y','y'],
                                                       return_signal=True)[0,...]
        #zzyy converges much faster than xxyy, but converges to the same result
        zzzzyy = [np.array([0,0,1])]*4 + [np.array([0,1,0])]*2
        sig_zy_manual_averaging = manual_average(self.TA6,zzzzyy,10)[0,:,:]
        abs_diff = np.max(np.abs(sig_zy_manual_averaging-sig_averaged_xy))
        rel_diff = abs_diff/np.max(np.abs(sig_averaged_xy))
        print(rel_diff)
        self.assertTrue(rel_diff<self.rel_tol)
if __name__ == '__main__':
    unittest.main()



"""
Module dipole_pruning
================
This module supplies one class, DipolePruning, that is used to find the
minimum number of elements required to resolve the dipole operator, to the
requested tolerance.
"""

import os

import numpy as np

class DipolePruning(object):
    r"""This class takes the dipole operator, mu, which must be expressed in
the eigenbasis of the system hamiltonian and uses Bessel's
inequality to determine the smallest number of states needed to
correctly resolve mu, to the given tolerance.  It expects a file
mu.npz in the folder 'file_path', with archive keys 'mu_GSM_to_SEM' and, 
optionally, 'mu_SEM_to_DEM'.  Each key must return a 3d numpy array with 
indices [i,j,k], where i and j are indices of the eigenvalues of the system 
hamiltonian, and k is an index 0,1,2 cooresponding to cartesian coordinates 
x,y,z.
"""
    def __init__(self,file_path):
        """Initialize object with
Args:
    file_path (str): file path to folder containing mu.npz
"""
        self.base_path = file_path
        self.load_mu()

    def load_mu(self):
        """Load the precalculated dipole overlaps.  The dipole operator must
be stored as a .npz file, and must contain a up to two arrays, each with 
three indices: (upper manifold eigenfunction, lower manifold eigenfunction, 
cartesian coordinate).  Keys: 'GSM_to_SEM' connects the ground state and 
singly excited manifolds, 'SEM_to_DEM' connects the singly and doubly excited
manifolds."""
        file_name = os.path.join(self.base_path,'mu.npz')
        mu = np.load(file_name)
        mu_keys = mu.keys()
        self.mu = {mu_key:mu[mu_key] for mu_key in mu_keys}

    def calculate_boolean_mu(self,overlap_matrix,*,rel_tol=1E-3):
        """Uses Bessel's inequality to find the minimum number of dipole
matrix elements needed to correctly resolve the dipole operator to the
given tolerance.

        Args:
            overlap_matrix (np.ndarray) : 3d-array of dipole matrix elements 
                [i,j,k] where i,j are eigenstates and k is a cartesian 
                coordinate.

            rel_tol (float) : relative tolerance for resolving the dipole
                operator mu.

"""
        dim0, dim1 = overlap_matrix.shape[:2]
        bool_mat = np.zeros((dim0,dim1),dtype=bool)
        
        # Inner product over cartesian coordinates: mu_ij dot mu_ij for each i,j pair
        # where mu_ij is a cartesian vector 
        prob_matrix = np.sum(overlap_matrix**2,axis=(2))

        # Sum over all lower manifold states
        probabilities = np.sum(prob_matrix,axis=1)

        # For each state n in the higher manifold
        for n in range(dim0):
            prob_tot = probabilities[n]
            # All lower states that connect to state n
            prob_list = prob_matrix[n,:]
            # Sort lower states by magnitude of mu_nj dot mu_nj
            prob_sort_ind = prob_list.argsort()[::-1]
            prob_sorted = prob_list[prob_sort_ind]
            prob = 0
            # Bessel's inequality
            for j in range(prob_sorted.size):
                prob += prob_sorted[j]
                if np.abs((prob_tot - prob)/prob_tot) < rel_tol:
                    # If the relative tolerance is attained, break out of loop
                    break
            #Keep only the states needed to satisfy the specified rel_tol
            non_zero_ind = prob_sort_ind[:j+1]

            #Set the states needed as True in a boolean array
            bool_mat[n,non_zero_ind] = True
        return bool_mat

    def save_boolean_mu(self,*,rel_tol = 1E-3):
        """Create and save the boolean masks for the dipole matrices
at the given tolerance. Files created by this function are mu_pruned.npz
and mu_boolean.npz.

        Args:
            rel_tol (float) : relative tolerance for resolving the dipole
                operator mu.  Default value of 0.001 has been found to work 
                well with vibronic systems to give convergence of the 
                Transient Absorption signal of better than 1%.
"""
        file_name_pruned = os.path.join(self.base_path,'mu_pruned.npz')
        file_name_boolean = os.path.join(self.base_path,'mu_boolean.npz')
        mu_GSM_to_SEM_boolean = self.calculate_boolean_mu(self.mu['GSM_to_SEM'],rel_tol=rel_tol)
        mu_GSM_to_SEM_pruned = self.mu['GSM_to_SEM'] * mu_GSM_to_SEM_boolean[:,:,np.newaxis]
        mu_boolean_dict = {'GSM_to_SEM':mu_GSM_to_SEM_boolean}
        mu_pruned_dict = {'GSM_to_SEM':mu_GSM_to_SEM_pruned}

        if 'SEM_to_DEM' in self.mu.keys():
            mu_SEM_to_DEM_boolean = self.calculate_boolean_mu(self.mu['SEM_to_DEM'],rel_tol=rel_tol)
            mu_SEM_to_DEM_pruned = self.mu['SEM_to_DEM'] * mu_SEM_to_DEM_boolean[:,:,np.newaxis]
            mu_boolean_dict['SEM_to_DEM'] = mu_SEM_to_DEM_boolean
            mu_pruned_dict['SEM_to_DEM'] = mu_SEM_to_DEM_pruned

        np.savez(file_name_pruned,**mu_pruned_dict)
        np.savez(file_name_boolean,**mu_boolean_dict)



#Standard python libraries
import os
import warnings
import time

#Dependencies - numpy, scipy
import numpy as np
from scipy.sparse import csr_matrix, identity, load_npz
from scipy.sparse.linalg import eigsh
from scipy.integrate import RK45

#Other parts of this code
from ufss.perturbative_calculations import DPBaseClass
from ufss.perturbative_calculations import ClosedBaseClass
from ufss.perturbative_calculations import RK_perturbative_container

class RKClosedEngine(ClosedBaseClass,DPBaseClass):
    """This class is designed to calculate perturbative wavepackets in the
        light-matter interaction given the hamiltonians of each optical
        manifold

    Args:
        file_path (string): path to folder containing eigenvalues and the
            dipole operator for the system Hamiltonian
        initial_state (int): index of initial state for psi^0 NOT IMPLEMENTED YET

"""
    def __init__(self,file_path,*,initial_state=0,
                 detection_type = 'polarization'):
        DPBaseClass.__init__(self,file_path,detection_type = detection_type)
        ClosedBaseClass.__init__(self,detection_type = detection_type)
        self.slicing_time = 0
        self.interpolation_time = 0
        self.expectation_time = 0
        self.next_order_expectation_time = 0
        self.convolution_time = 0
        self.extend_time = 0
        self.mask_time = 0
        self.dipole_time = 0
        self.automation_time = 0
        self.diagram_to_signal_time = 0
        self.RK45_step_time = 0

        self.sparsity_threshold = 0.1

        self.interaction_picture_calculations = True

        self.load_mu()

        self.load_H()

        self.set_min_max_manifolds()

        self.rtol = 1E-6
        self.atol = 1E-6

        self.time_to_extend = 0
        self.time_for_next_order = 0

    def check_sparsity(self,mat):
        csr_mat = csr_matrix(mat)
        sparsity = csr_mat.nnz / (csr_mat.shape[0]*csr_mat.shape[1])
        if sparsity < self.sparsity_threshold:
            return True
        else:
            return False

    def set_efields(self,times_list,efields_list,centers_list,
                    phase_discrimination,*,reset_calculations = True,
                    plot_fields = False):

        DPBaseClass.set_efields(self,times_list,efields_list,centers_list,
                                 phase_discrimination,
                                 reset_calculations = reset_calculations,
                                 plot_fields = plot_fields)
        
        # Initialize unperturbed density matrix
        self.set_psi0_auto()

    def add_impulsive_psis(self,a,b):
        """Add two perturbative_containers together
"""
        manifold_key = a.manifold_key
        pdc = a.pdc
        sim_a = a.simultaneous
        sim_b = b.simultaneous
        if sim_a != sim_b:
            raise Exception('Cannot add impulsive psis if they do not have the same number of simultaneous interactions')
        else:
            simultaneous = sim_a
        pna = a.pulse_number
        pnb = a.pulse_number
        ta = self.pulse_times[pna]
        tb = self.pulse_times[pnb]
        if np.allclose(ta,tb):
            t = a.t[1:2]
            psi_a = a.f[:,1:2]
            psi_b = b.f[:,1:2]
            pulse_number = max(pna,pnb)
        else:
            raise Exception('Cannot add two impulsive psis unless they have the same value of t0')

        psi = psi_a + psi_b
        psi *= 2 * simultaneous
        
        ab=RK_perturbative_container(t,psi,pulse_number,manifold_key,pdc,
                                     simultaneous = simultaneous)
        ab.one_time_step_function = self.one_time_step_function
        return ab
        

    def add_psis(self,a,b):
        """Add two density matrix objects together
"""
        if not np.allclose(a.pdc,b.pdc):
            raise Exception('Cannot add density matrices with different phase-discrimination conditions')
        if 'all_manifolds' in self.manifolds:
            pass
        else:
            if a.manifold_key != b.manifold_key:
                raise Exception('Cannot add perturbative_container objects that exist in different manifolds')
            
        if a.impulsive:
            return self.add_impulsive_psis(a,b)
        
        manifold_key = a.manifold_key
        
        pdc = a.pdc
        
        tmin = min(a.t[0],b.t[0])
        tmax = max(a.t[-1],b.t[-1])
        dt = min(a.t[1]-a.t[0],b.t[1]-b.t[0])
        t = np.arange(tmin,tmax+dt*0.9,dt)
        psi = a(t) + b(t)

        pna = a.pulse_number
        pnb = b.pulse_number
        pulse_number = max(pna,pnb)
        
        ab = RK_perturbative_container(t,psi,pulse_number,manifold_key,pdc)
        ab.one_time_step_function = self.one_time_step_function
        return ab

    def get_bottom_eigenvector(self):
        try:
            H = self.H['all_manifolds']
        except KeyError:
            H = self.H['0']
        if H.shape == (1,1):
            e = H[0,0]
            ev = np.array([[1]])
        else:
            e, ev = eigsh(H,k=1,which='SM',maxiter=10000)

        for key in self.H.keys():
            sh = self.H[key].shape
            if type(self.H[key]) is np.ndarray:
                self.H[key] -= e * np.eye(sh[0])
            else:
                self.H[key] -= e * identity(sh[0])

        v = ev[:,0]
        return e, v

    def set_psi0_auto(self):
        try:
            psi0 = np.load(os.path.join(self.base_path,'psi0.npy'))
        except FileNotFoundError:
            e0, psi0 = self.get_bottom_eigenvector()
        self.set_psi0_manual(psi0)

    def set_psi0_manual(self,psi0):
        t = np.array([-np.inf,0,np.inf])
        psi0 = psi0[:,np.newaxis] * np.ones((psi0.size,t.size))
        pulse_number = None
        manifold_key = self.ordered_manifolds[0]
        pdc = np.zeros(self.pdc.shape,dtype=self.pdc.dtype)
        
        self.psi0 = RK_perturbative_container(t,psi0,pulse_number,
                                              manifold_key,pdc,
                                              interp_kind = 'zero')

    def load_mu(self):
        """Load the precalculated dipole overlaps.  The dipole operator must
            be stored as a .npz file, and must contain at least one array, each with three 
            indices: (new manifold index, old manifold eigenfunction, 
            cartesian coordinate)."""
        try:
            file_name = os.path.join(self.base_path,'mu_site_basis.npz')
            with np.load(file_name) as mu_archive:
                self.H_mu={key:mu_archive[key] for key in mu_archive.keys()}
        except FileNotFoundError:
            try:
                file_name = os.path.join(self.base_path,'mu_original_H_basis.npz')
                with np.load(file_name) as mu_archive:
                    self.H_mu = {key:mu_archive[key] for key
                                 in mu_archive.keys()}
            except FileNotFoundError:
                file_name = os.path.join(self.base_path,'mu.npz')
                with np.load(file_name) as mu_archive:
                    self.H_mu = {key:mu_archive[key] for key
                                 in mu_archive.keys()}
        sparse_flags = []
        for key in self.H_mu.keys():
            mu_2D = np.sum(np.abs(self.H_mu[key])**2,axis=-1)
            sparse_flags.append(self.check_sparsity(mu_2D))
        sparse_flags = np.array(sparse_flags)
        if np.allclose(sparse_flags,True):
            self.sparse_mu_flag = True
        else:
            self.sparse_mu_flag = False

        for key in self.H_mu.keys():
            mu_x = self.H_mu[key][...,0]
            mu_y = self.H_mu[key][...,1]
            mu_z = self.H_mu[key][...,2]

            if self.sparse_mu_flag:
                self.H_mu[key] = [csr_matrix(mu_x),csr_matrix(mu_y),
                                  csr_matrix(mu_z)]
            else:
                self.H_mu[key] = [mu_x,mu_y,mu_z]

        print('RKE_sparse_mu_flag',self.sparse_mu_flag)

    def load_H(self):
        """Load in known eigenvalues. Must be stored as a numpy archive file,
with keys: GSM, SEM, and optionally DEM.  The eigenvalues for each manifold
must be 1d arrays, and are assumed to be ordered by increasing energy. The
energy difference between the lowest energy ground state and the lowest 
energy singly-excited state should be set to 0
"""
        H_file_name = os.path.join(self.base_path,'H.npz')
        try:
            with np.load(H_file_name,allow_pickle=True) as H_archive:
                self.H = dict()
                for key in H_archive.keys():
                    H = H_archive[key]
                    if H.dtype == np.dtype('O'):
                        self.H[key] = H[()]
                    else:
                        if self.check_sparsity(H):
                            self.H[key] = csr_matrix(H)
                        else:
                            self.H[key] = H
        except:
            self.H = {'all_manifolds':load_npz(H_file_name)}
        self.manifolds = list(self.H.keys())

        if '0' in self.manifolds:
            self.ordered_manifolds = [str(i) for i in range(len(self.manifolds))]
        elif 'all_manifolds' in self.manifolds:
            self.ordered_manifolds = ['all_manifolds']
        else:
            self.ordered_manifolds = ['GSM','SEM','DEM','TEM','QEM']

    def dH(self,t,psi):
        try:
            H = self.H['all_manifolds']
        except KeyError:
            H = self.H[psi.manifold_key]
        return -1j*H.dot(psi)

    def get_dH_manual(self,manifold_key):
        try:
            H = self.H['all_manifolds']
        except KeyError:
            H = self.H[manifold_key]

        def H_fun(t,psi):
            return -1j*H.dot(psi)

        return H_fun

    def one_time_step_function(self,psi0,t0,tf,*,manifold_key = None):
        num_steps = 0
        if manifold_key == None:
            rk45 = RK45(self.dH,t0,psi0,tf,atol=self.atol,rtol=self.rtol)
        else:
            dH = self.get_dH_manual(manifold_key)
            rk45 = RK45(dH,t0,psi0,tf,atol=self.atol,rtol=self.rtol)
        if rk45.direction == 1:
            while rk45.t < tf:
                rk45.step()
                num_steps += 1
        else:
            while rk45.t > tf:
                rk45.step()
                num_steps += 1
        psi_final = rk45.y
        
        return psi_final


    def mu_dot_psi(self,t,psi_in,pulse_number,up_flag=True,
                   next_manifold_mask = None):
        old_manifold_key = psi_in.manifold_key
        new_manifold_key=self.get_new_manifold_key(old_manifold_key,up_flag)

        if up_flag:
            mu_key = old_manifold_key + '_to_' + new_manifold_key
        else:
            mu_key = new_manifold_key + '_to_' + old_manifold_key

        mu = self.get_H_mu(pulse_number,mu_key,rotating_flag=up_flag)
            
        psi = self.get_psi(t,psi_in)

        if psi.size == 0:
            psi = np.array([])
        else:
            psi = mu.dot(psi)
            
        return psi

    def mu_exp_val(self,t,psis_in,pulse_number,*,mu_type = None,
                   ket_flag = None, up_flag = None):
        """Keyword arguments are provided simply to make compatible with 
            open system framework
"""
        bra_in, ket_in = psis_in
        bra_key = bra_in.manifold_key
        ket_key = ket_in.manifold_key

        bra_num = self.manifold_key_to_number(bra_key)
        ket_num = self.manifold_key_to_number(ket_key)

        if 'all_manifolds' in self.manifolds:
            pass
        elif np.abs(bra_num - ket_num) != 1:
            warnings.warn('Dipole only connects manifolds 0 to 1 or 1 to 2')
            return None
        t0 = time.time()
        if bra_num > ket_num:
            bra = self.mu_dot_psi(t,bra_in,-1,up_flag=False)
            ket = self.get_psi(t,ket_in)
        else:
            ket = self.mu_dot_psi(t,ket_in,-1,up_flag=False)
            bra = self.get_psi(t,bra_in)

        exp_val = np.sum(np.conjugate(bra) * ket,axis = 0)
            
        return exp_val

    def next_order(self,psi_in,*,up_flag=True,
                   next_manifold_mask = None,pulse_number = 0):
        """This function connects psi_p to psi_pj^(*) using a DFT convolution algorithm.

        Args:
            psi_in (psi_container): input density matrix
            pulse_number (int): index of optical pulse (0,1,2,...) can also be set to
                'impulsive'
            next_manifold_mask (np.ndarray): optional - define the states to be considered 
                in the next manifold
        
        Return:
            psi_dict (psi_container): next-order wavefunction
"""
        # Rotating term excites the ket
        conjugate_flag = not up_flag

        if psi_in.pulse_number == None:
            old_pulse_time = -np.inf
        else:
            old_pulse_time = self.pulse_times[psi_in.pulse_number]
        pulse_time = self.pulse_times[pulse_number]

        if np.isclose(old_pulse_time,pulse_time):
            simultaneous = psi_in.simultaneous + 1
        else:
            simultaneous = 1

        t = self.efield_times[pulse_number] + pulse_time
        dt = self.dts[pulse_number]
        old_manifold_key = psi_in.manifold_key
        new_manifold_key=self.get_new_manifold_key(old_manifold_key,up_flag)

        input_pdc = psi_in.pdc
        output_pdc = input_pdc.copy()
        if conjugate_flag:
            output_pdc[pulse_number][1] += 1
        else:
            output_pdc[pulse_number][0] += 1

        center = self.centers[pulse_number]
        efield = self.efields[pulse_number] * np.exp(-1j*center*t)

        if conjugate_flag:
            efield = np.conjugate(efield)

        # factor of 1j comes from perturbation theory
        efield = 1j * efield
        
        t0 = time.time()
        psi = self.mu_dot_psi(t,psi_in,pulse_number,up_flag=up_flag,
                              next_manifold_mask=next_manifold_mask)
        
        t1 = time.time()
        self.next_order_expectation_time += t1-t0

        psi = psi * efield[np.newaxis,:]

        M = t.size

        if M == 1:
            next_psi = psi
        else:
            next_psi = np.zeros(psi.shape,dtype='complex')

            dt = self.dts[pulse_number]
            
            if self.method == 'Euler':
                next_psi[:,0] = psi[:,0] * dt
                for i in range(1,t.size):
                    psi0 = next_psi[:,i-1]
                    t0 = t[i-1]
                    ta = time.time()
                    next_psi[:,i] = self.one_time_step_function(psi0,t0,t[i],manifold_key=new_manifold_key)

                    tb = time.time()
                    self.RK45_step_time += tb - ta

                    next_psi[:,i] += psi[:,i] * dt

            elif self.method == 'UF2':
                ref_time = pulse_time
                for i in range(t.size):
                    next_psi[:,i]=self.one_time_step_function(psi[:,i],t[i],
                                    ref_time,manifold_key=new_manifold_key)

                t0 = time.time()

                M = self.efield_times[pulse_number].size

                fft_convolve_fun = self.heaviside_convolve_list[pulse_number].fft_convolve2

                next_psi = fft_convolve_fun(next_psi,d=dt)

                t1 = time.time()
                self.convolution_time += t1-t0

                for i in range(t.size):
                    next_psi[:,i]=self.one_time_step_function(next_psi[:,i],
                                ref_time,t[i],manifold_key=new_manifold_key)

        psi_out = RK_perturbative_container(t,next_psi,pulse_number,
                                            new_manifold_key,output_pdc,
                                            simultaneous=simultaneous)
        psi_out.one_time_step_function = self.one_time_step_function
    
        return psi_out

    def get_psi(self,t,psi_obj):
        psi = psi_obj(t)
        return psi



import os
import time
import yaml

import numpy as np
from scipy.sparse import csr_matrix

from ufss import CompositeDiagrams

class OpenBaseClass(CompositeDiagrams):
    def __init__(self,*,detection_type='polarization'):
        CompositeDiagrams.__init__(self,detection_type=detection_type)

        self.number_of_diagrams_calculated = 0
        
        self.KB_dict = {'Bu':self.bra_up,'Ku':self.ket_up,
                        'Kd':self.ket_down,'Bd':self.bra_down}

        self.rhos = dict()
        self.composite_rhos = dict()

    def load_H_eigensystem(self):
        """Look for the eigenvalues and eigenvectors of H, and load them 
            if available. Useful as a convenience for assigning spectral 
            positions"""
        parent_dir = os.path.split(self.base_path)[0]
        H_e_name = os.path.join(parent_dir,'closed','eigenvalues.npz')
        H_ev_name = os.path.join(parent_dir,'closed','eigenvectors.npz')
        try:
            with np.load(H_e_name) as eigval_archive:
                manifolds = list(eigval_archive.keys())
                self.H_eigenvalues = {key:eigval_archive[key] for key in manifolds}
        except FileNotFoundError:
            pass
        try:
            with np.load(H_ev_name) as eigvec_archive:
                manifolds = list(eigvec_archive.keys())
                self.H_eigenvectors = {key:eigvec_archive[key] for key in manifolds}
        except FileNotFoundError:
            pass

    def load_fl_yield(self):
        parent_dir = os.path.split(self.base_path)[0]
        file_name = os.path.join(parent_dir,'closed','fluorescence_yield.npz')

        with np.load(file_name) as fl_archive:
            self.fluorescence_yield = {key:fl_archive[key] for key in fl_archive.keys()}

    ### common method for loading and using the correct mu
    def load_H_mu(self):
        parent_dir = os.path.split(self.base_path)[0]
        params_file = os.path.join(parent_dir,'params.yaml')
        with open(params_file) as yamlstream:
            params = yaml.load(yamlstream,Loader=yaml.SafeLoader)
        if 'ManualD' in params.keys():
            if params['H_eigenbasis']:
                file_name = os.path.join(parent_dir,'closed','mu.npz')
            else:
                file_name = os.path.join(parent_dir,'closed',
                                     'mu_original_H_basis.npz')
        elif ('site_bath' in params['bath'].keys() or 
            'vibration_bath' in params['bath'].keys()):
            file_name = os.path.join(parent_dir,'closed','mu.npz')
        else:
            file_name = os.path.join(parent_dir,'closed',
                                     'mu_original_H_basis.npz')

        with np.load(file_name) as mu_archive:
            self.H_mu = {key:mu_archive[key] for key in mu_archive.keys()}

    def get_H_mu(self,pulse_number,key,rotating_flag=True):
        """Calculates the dipole matrix given the electric field 
            polarization vector"""
        t0 = time.time()
        pol = self.polarization_sequence[pulse_number]

        x = np.array([1,0,0])
        y = np.array([0,1,0])
        z = np.array([0,0,1])
        try:
            mu = self.H_mu[key]
        except KeyError:
            try:
                key = 'up'
                mu = self.H_mu[key]
            except KeyError:
                key = 'ket_up'
                mu = self.H_mu[key]
            
        if np.all(pol == x):
            overlap_matrix = mu[:,:,0]#.copy()
        elif np.all(pol == y):
            overlap_matrix = mu[:,:,1]#.copy()
        elif np.all(pol == z):
            overlap_matrix = mu[:,:,2]#.copy()
        else:
            overlap_matrix = np.tensordot(mu,pol,axes=(-1,0))

        if not rotating_flag:
            overlap_matrix = overlap_matrix.T

        t1 = time.time()
        self.dipole_time += t1-t0

        return overlap_matrix

    def dipole_matrix(self,pulse_number,key,ket_flag=True,up_flag=True):
        """Calculates the dipole matrix given the electric field polarization vector,
            if ket_flag = False then uses the bra-interaction"""
        t0 = time.time()
        pol = self.polarization_sequence[pulse_number]
            
        x = np.array([1,0,0])
        y = np.array([0,1,0])
        z = np.array([0,0,1])
        try:
            mu = self.mu[key]
        except KeyError:
            if ket_flag:
                key = 'ket'
            else:
                key = 'bra'
            if up_flag:
                key += '_up'
            else:
                key += '_down'
            mu = self.mu[key]
            
        if np.all(pol == x):
            overlap_matrix = mu[0]#.copy()
        elif np.all(pol == y):
            overlap_matrix = mu[1]#.copy()
        elif np.all(pol == z):
            overlap_matrix = mu[2]#.copy()
        else:
            overlap_matrix = mu[0]*pol[0] + mu[1]*pol[1] + mu[2]*pol[2]

        t1 = time.time()
        self.dipole_time += t1-t0

        return overlap_matrix

    def get_mu(self,pulse_number,old_manifold_key,new_manifold_key,
               ket_flag=True,up_flag=True,mu_type='L'):
        if mu_type=='H':
            old_k,old_b = old_manifold_key.split(',')
            new_k,new_b = new_manifold_key.split(',')

            if ket_flag:
                if up_flag:
                    H_mu_key = old_k + '_to_' + new_k
                else:
                    H_mu_key = new_k + '_to_' + old_k

                rotating_flag = up_flag
            else:
                if up_flag:
                    H_mu_key = old_b + '_to_' + new_b
                else:
                    H_mu_key = new_b + '_to_' + old_b
                rotating_flag = not up_flag

            overlap_matrix = self.get_H_mu(pulse_number,
                                H_mu_key,rotating_flag=rotating_flag)
        else:
            mu_key = old_manifold_key + '_to_' + new_manifold_key
            overlap_matrix = self.dipole_matrix(pulse_number,mu_key,
                                        ket_flag=ket_flag,up_flag=up_flag)

        return overlap_matrix

    ### Generic methods for calculating individual diagrams

    def execute_diagram(self,instructions):
        self.number_of_diagrams_calculated += 1
        r = self.rho0
        name = ''
        for i in range(len(instructions)):
            key, num = instructions[i]
            name += key+str(num)
            # Try to re-use previous calculations, if they exist
            try:
                new_r = self.rhos[name]
            except KeyError:
                new_r = self.KB_dict[key](r,pulse_number=num)
                self.rhos[name] = new_r
            r = new_r
            
        return r

    def execute_composite_diagrams(self):
        rhos = self.composite_rhos
        old_rhos = [self.rho0]
            
        for i in range(self.highest_order):
            new_rhos = []
            pdcs = set()
            for old_rho in old_rhos:
                old_pdc = old_rho.pdc_tuple
                next_instructions = self.get_next_interactions(old_pdc)
                for key in next_instructions.keys():
                    pdcs.add(key)
                    if key in rhos:
                        # do not do calculations if they are already done
                        pass
                    else:
                        ket_ins,bra_ins = next_instructions[key]
                        new_rho_k =self.execute_interaction(old_rho,ket_ins)
                        new_rho_b =self.execute_interaction(old_rho,bra_ins)
                        new_rhos.append(new_rho_k)
                        new_rhos.append(new_rho_b)
            
            for new_rho in new_rhos:
                if new_rho.pdc_tuple in rhos.keys():
                    partial_rho = rhos[new_rho.pdc_tuple]
                    new_rho = self.add_rhos(partial_rho,new_rho)
                rhos[new_rho.pdc_tuple] = new_rho

            old_rhos = [rhos[pdc] for pdc in pdcs]

        return None

    def remove_calculations_by_pulse_number(self,pulse_number):
        num = str(pulse_number)
        keys = self.rhos.keys()
        keys_to_remove = []
        for key in keys:
            flag = key.find(num)
            if flag >= 0:
                keys_to_remove.append(key)
        for key in keys_to_remove:
            self.rhos.pop(key)

    def remove_composite_calculations_by_pulse_number(self,pulse_number):
        keys = self.composite_rhos.keys()
        keys_to_remove = []
        for key in keys:
            ket_ints,bra_ints = key[pulse_number]
            if ket_ints > 0 or bra_ints > 0:
                keys_to_remove.append(key)
        for key in keys_to_remove:
            self.composite_rhos.pop(key)

    def set_current_diagram_instructions(self,arrival_times):
        self.diagram_generation_counter += 1
        t0a = time.time()
        self.current_instructions = self.get_diagram_dict(arrival_times)
        t0b = time.time()
        self.diagram_generation_time = t0b - t0a

    def reset(self):
        self.rhos = dict()
        self.composite_rhos = dict()

    def convert_mu_keys(self):
        old_keys = list(self.mu.keys())
        print(old_keys)
        for key in old_keys:
            key1,key2 = key.split('_to_')
            old_k,old_b = key1
            new_k,new_b = key2
            new_key = ''.join([old_k,',',old_b,'_to_',new_k,',',new_b])
            self.mu[new_key] = self.mu.pop(key)
            self.mu_boolean[new_key] = self.mu_boolean.pop(key)

    def set_maximum_manifold(self,num='auto'):
        
        if num == 'auto':
            if 'all_manifolds' in self.manifolds:
                self.maximum_manifold = np.inf
            else:
                manifold_numbers = []
                for key in self.manifolds:
                    a,b = key.split(',')
                    if a == b:
                        manifold_numbers.append(int(a))
                self.maximum_manifold = max(manifold_numbers)
        else:
            self.maximum_manifold = num

    def manifold_array_to_key(self,manifold):
        """Inverse of self.manifold_key_to_array"""
        if manifold.size != 2 or manifold.dtype != int:
            raise Exception('manifold array must contain exactly 2 integer') 
        return str(manifold[0]) + ',' + str(manifold[1])

    def manifold_key_to_array(self,key):
        """Key must be a string containing two integers separated by a comma, 
            the first describing
            the ket manifold, the second the bra manifold.  If the density 
            matrix is represented in the full space, rather than being divided
            into manifolds, the first integer reperesents the total number of
            excitations to the ket side, and the second integer represents 
            the sum of all excitations to the bra side."""
        return np.array([int(char) for char in key.split(',')],dtype=int)

    def execute_interaction(self,rho_in,interaction):
        """This method connects the input density matrix of order n to a
            density matrix that contributes to order n + 1, according to the
            specified interaction type and pulse number
        Args:
            rho_in (DensityMatrices) : input density matrix
            interaction (tuple) : specifies interaction type and pulse 
                number in format (str, int)
"""
        int_type, pulse_num = interaction
        return self.KB_dict[int_type](rho_in,pulse_number = pulse_num)
            
    def ket_up(self,rho_in,*,next_manifold_mask = None,pulse_number = 0):
        """This method connects psi_p to psi_pj where the next order psi
            is one manifold above the current manifold.

        Args:
            rho_in (perturbative_container): input density matrix
            pulse_number (int): index of optical pulse (0,1,2,...) 
            next_manifold_mask (np.ndarray): optional - define the states to be considered 
                in the next manifold

        Returns:
            (perturbative_container): output from method next_order
"""

        return self.next_order(rho_in,ket_flag=True,up_flag=True,
                               next_manifold_mask = next_manifold_mask,
                               pulse_number = pulse_number)

    def ket_down(self,rho_in,*,next_manifold_mask = None,pulse_number = 0):
        """This method connects psi_p to psi_pj where the next order psi
            is one manifold above the current manifold.

        Args:
            rho_in (perturbative_container): input density matrix
            pulse_number (int): index of optical pulse (0,1,2,...) 
            next_manifold_mask (np.ndarray): optional - define the states to be considered 
                in the next manifold

        Returns:
            (perturbative_container): output from method next_order
"""

        return self.next_order(rho_in,ket_flag=True,up_flag=False,
                               next_manifold_mask = next_manifold_mask,
                               pulse_number = pulse_number)

    def bra_up(self,rho_in,*,next_manifold_mask = None,pulse_number = 0):
        """This method connects psi_p to psi_pj where the next order psi
            is one manifold above the current manifold.

        Args:
            rho_in (perturbative_container): input density matrix
            pulse_number (int): index of optical pulse (0,1,2,...)
            next_manifold_mask (np.ndarray): optional - define the states to be considered 
                in the next manifold

        Returns:
            (perturbative_container): output from method next_order
"""

        return self.next_order(rho_in,ket_flag=False,up_flag=True,
                               next_manifold_mask = next_manifold_mask,
                               pulse_number = pulse_number)

    def bra_down(self,rho_in,*,next_manifold_mask = None,pulse_number = 0):
        """This method connects psi_p to psi_pj where the next order psi
            is one manifold above the current manifold.

        Args:
            rho_in (perturbative_container): input density matrix
            pulse_number (int): index of optical pulse (0,1,2,...)
            next_manifold_mask (np.ndarray): optional - define the states to be considered 
                in the next manifold

        Returns:
            (perturbative_container): output from method next_order
"""

        return self.next_order(rho_in,ket_flag=False,up_flag=False,
                               next_manifold_mask = next_manifold_mask,
                               pulse_number = pulse_number)

    def get_rho_by_key(self,t,key,*,reshape=True,original_L_basis=True):
        try:
            rho = self.composite_rhos[key]
        except KeyError:
            rho = self.rhos[key]
        return self.get_rho(t,rho,reshape=reshape,
                            original_L_basis=original_L_basis)

    def get_rho_by_order(self,t,order,*,reshape=True,original_L_basis=True):
        if len(self.composite_rhos) > 0:
            keys = self.composite_rhos.keys()
            order_keys = []
            for key in keys:
                pdc = self.pdc_tup_to_arr(key)
                if pdc.sum() == order:
                    order_keys.append(key)
        else:
            keys = self.rhos.keys()
            order_keys = []
            for key in keys:
                if len(key) == 3*order:
                    order_keys.append(key)
        
        rho_total = self.get_rho_by_key(t,order_keys.pop(0),reshape=reshape,
                                        original_L_basis=original_L_basis)
        for key in order_keys:
            rho_total += self.get_rho_by_key(t,key,reshape=reshape,
                                        original_L_basis=original_L_basis)

        return rho_total

    def check_sparsity(self,mat):
        csr_mat = csr_matrix(mat)
        sparsity = csr_mat.nnz / (csr_mat.shape[0]*csr_mat.shape[1])
        if sparsity < self.sparsity_threshold:
            return True
        else:
            return False

    def get_new_manifold_key(self,old_manifold_key,ket_flag,up_flag):
        
        if up_flag:
            change = 1
        else:
            change = -1
        if ket_flag:
            manifold_change = np.array([change,0],dtype=int)
        else:
            manifold_change = np.array([0,change],dtype=int)
        old_manifold = self.manifold_key_to_array(old_manifold_key)
        new_manifold = old_manifold + manifold_change
        new_manifold_key = self.manifold_array_to_key(new_manifold)
        
        return new_manifold_key



import unittest
import numpy as np
import os
import yaml
from scipy.io import loadmat

import ufss

folder = 'fixtures/dimer_d0.2'

def make_L():
    site_energies = [0,0.75]
    site_couplings = [0.3307]
    dipoles = [[1,0,0],[0,1,0]]

    site_bath = {'cutoff_frequency':1,
                 'coupling':0.05,
                 'temperature':1,
                 'cutoff_function':'lorentz-drude',
                 'spectrum_type':'ohmic'}

    vibration_bath = site_bath
    trunc_size = 1

    vibraations = []

    vibrations = [{'displacement':0,'site_label':0,'omega_g':1},
                  {'displacement':0,'site_label':1,'omega_g':1.01}]

    os.makedirs(folder,exist_ok=True)


    secular_bath = {'secular':True,
            'site_bath':site_bath,
            'vibration_bath':vibration_bath}

    secular_params = {
        'site_energies':site_energies,
        'site_couplings':site_couplings,
        'dipoles':dipoles,
        'truncation_size':trunc_size,
        'num_eigenvalues':'full',
        'eigenvalue_precision':1,
        'vibrations':vibrations,
        'bath':secular_bath}

    with open(os.path.join(folder,'simple_params.yaml'),'w+') as new_file:
        yaml.dump(secular_params,new_file)

    ufss.HLG.run(folder,conserve_memory=False)

    return None

def gaussian(t,sigma):
    """t is time.  Gaussian pulse, with time-domain standard deviation sigma, 
    normalized to behave like a delta function as sigma -> 0"""
    pre = 1/(np.sqrt(2*np.pi)*sigma)
    return pre * np.exp(-t**2/(2*sigma**2))

def setup(uf2_flag=True,conserve_memory=True,open=True):
    open_folder = os.path.join(folder,'open')
    closed_folder = os.path.join(folder,'closed')
    if open:
        if uf2_flag:
            obj = ufss.DensityMatrices(open_folder,detection_type='polarization',conserve_memory=conserve_memory)
            M = 41
        else:
            obj = ufss.RKE_DensityMatrices(open_folder,detection_type='polarization',conserve_memory=conserve_memory)
            M = 801
    else:
        if uf2_flag:
            obj = ufss.Wavepackets(closed_folder,detection_type='polarization')
            M = 41
        else:
            obj = ufss.RKWavefunctions(closed_folder,detection_type='polarization')
            M = 801


    sigma = 1
    Delta = 10
    t = np.linspace(-Delta/2,Delta/2,num=M)*sigma
    ef = gaussian(t,sigma)

    obj.set_efields([t,t,t,t],[ef,ef,ef,ef],[0,0,0,0],[(0,1),(1,0),(1,0)])

    obj.set_polarization_sequence(['x','x','x','x'])
    obj.set_t(0.137,dt=0.5)

    tau = np.array([0])

    T = np.arange(10)

    obj.set_pulse_delays([tau,T])

    return obj

def L2_norm(a,b):
    """Computes L2 norm of two nd-arrays, taking the b as the reference"""
    return np.sqrt(np.sum(np.abs(a-b)**2)/np.sum(np.abs(b)**2))
    

class test_UF2_vs_RKE(unittest.TestCase):
    """This is to test that UF2 and RKE gives the same result, to
        within the expected tolerance. This tests the UF2 and RKE
        modules, both for open and closed calculations, but does
        not test the HLG, since both UF2 and RKE are relying on it
        for this test.
"""
    def test_1(self):
        make_L()
        uf2 = setup(uf2_flag = True)
        uf2_sig = uf2.calculate_signal_all_delays()

        rke = setup(uf2_flag = False)
        rke_sig = rke.calculate_signal_all_delays()

        diff = L2_norm(uf2_sig,rke_sig)
        self.assertTrue(diff < 0.01)

    def test_2(self):
        make_L()
        uf2 = setup(uf2_flag = True,conserve_memory=False)
        uf2_sig = uf2.calculate_signal_all_delays()

        rke = setup(uf2_flag = False,conserve_memory=False)
        rke_sig = rke.calculate_signal_all_delays()

        diff = L2_norm(uf2_sig,rke_sig)
        self.assertTrue(diff < 0.01)

    def test_3(self):
        make_L()
        uf2 = setup(uf2_flag = True, open=False)
        uf2_sig = uf2.calculate_signal_all_delays()

        rke = setup(uf2_flag = False, open=False)
        rke_sig = rke.calculate_signal_all_delays()

        diff = L2_norm(uf2_sig,rke_sig)
        self.assertTrue(diff < 0.01)

if __name__ == '__main__':
    unittest.main()



from .calculate_signals import UF2DensityMatrices, UF2Wavefunctions, RKDensityMatrices, RKWavefunctions, SpectroscopyBase
from .plotting_tools import Plotter, plot2D, plot2D_logscaled
from .signal_processing import SignalProcessing
from .isotropic_averaging import IsotropicAverage, FWMIsotropicAverage
from .analyze_TA_example import AnalyzeTransientAbsorption
from .tdes import PumpPumpProbe, Rephasing, NonRephasing
from .transient_absorption_signals import TransientAbsorption



import numpy as np
import matplotlib.pyplot as plt
import numpy.polynomial.chebyshev as npch
try:
    from pyfftw.interfaces.numpy_fft import fft, fftshift, ifft, ifftshift, fftfreq
    from ufss.perturbative_calculations.heaviside_convolve_pyfftw import HeavisideConvolve, HeavisideConvolveSP
except ImportError:
    from numpy.fft import fft, fftshift, ifft, ifftshift, fftfreq
    from ufss.perturbative_calculations.heaviside_convolve_numpy import HeavisideConvolve, HeavisideConvolveSP
import copy
import warnings
import os

class BaseClass:
    """Contains methods that are used by all perturbative calculation engines
        (both UF2 and direct propagation).
"""
    def __init__(self,file_path,*,detection_type = 'polarization'):
        self.detection_type = detection_type
        self.base_path = file_path
        self.method = 'UF2'

        # Code will not actually function until the following empty lists are set by the user
        self.efields = [] #initialize empty list of electric field shapes
        self.efield_times = [] #initialize empty list of times assoicated with each electric field shape
        self.dts = [] #initialize empty list of time spacings associated with each electric field shape
        self.polarization_sequence = [] #initialize empty polarization sequence
        self.pulse_times = [] #initialize empty list of pulse arrival times
        self.centers = [] #initialize empty list of pulse center frequencies
        self.efield_wavevectors = []
        self.heaviside_convolve_list = []

    def set_efields(self,times_list,efields_list,centers_list,
                    phase_discrimination,*,reset_calculations = True,
                    plot_fields = False):
        """Use to set electric field pulse shapes to be used in calculations. 
            Electric fields can include the carrier frequency, in which case the
            centers in centers_list should be 0, or can be specified in the 
            rotating frame, in which case center frequencies should be specified
            in the centers_list.
            
        Args:
            times_list (list) : list of 1D np.ndarrays specifying the time points
                on which each pulse is defined.
            efields_list (list) : list of 1D np.ndarrays specifying the complex 
                amplitude of each pulse
            centers_list (list) : list of floats specifying the center frequency
                of each pulse
            phase_discrimination (tuple) : tuple of tuples, ((n0r,n0c),(n1r,n1c),...)
                that specify the number of interactions with the rotating and
                counter-rotating terms of each pulse
                
        Keyword Args:
            reset_calculations (bool) : True if all stored perturbative calculations
                should be deleted (setting this to False is dangerous, as old 
                calculations done with previous pulse-shapes will be reused without
                warning)
            plot_fields (bool) : True if you wish to see plots of the pulses in both
                time- and frequency-domain. This will also initiate a check of the
                time- and frequency-resolution of the pulses
            """
        self.efield_times = times_list
        self.efields = efields_list
        self.efield_masks = [dict() for efield in self.efields]
        self.centers = centers_list
        self.set_phase_discrimination(phase_discrimination)
        self.dts = []
        self.efield_frequencies = []
        self.heaviside_convolve_list = []
        if reset_calculations:
            self.reset()
        for t in times_list:
            if t.size == 1:
                dt = 1
                w = np.array([0])
            else:
                dt = t[1] - t[0]
                w = fftshift(fftfreq(t.size,d=dt))*2*np.pi
            self.dts.append(dt)
            self.efield_frequencies.append(w)
            if self.interaction_picture_calculations:
                self.heaviside_convolve_list.append(HeavisideConvolve(t.size))
            else:
                self.heaviside_convolve_list.append(HeavisideConvolveSP(t.size))
        self.dt = self.dts[0]
        self.set_dft_efields()

        if self.detection_type == 'polarization' or 'integrated_polarization':
            try:
                self.local_oscillator = self.efields[-1].copy()
            except:
                self.local_oscillator = copy.deepcopy(self.efields[-1])

        for i in range(len(self.efields)):
            self.check_efield_resolution(i,plot_fields = plot_fields)

    def set_dft_efields(self):
        """Calculates the dft of the efields and saves the results as a list 
            to the attribute "efields_fts". Also calculates the associated 
            angular frequencies, and stores the results as a list to the
            attribute "efield_omegas" """
        self.efield_omegas = []
        self.efield_fts = []
        for pulse_number in range(len(self.efields)):
            dt = self.dts[pulse_number]
            efield_t = self.efield_times[pulse_number]
            efield_w = fftshift(fftfreq(efield_t.size,d=dt))*2*np.pi
            self.efield_omegas.append(efield_w)
            ifft_norm = dt*efield_t.size
            efield = self.efields[pulse_number]
            efield_ft = fftshift(ifft(ifftshift(efield)))*ifft_norm
            self.efield_fts.append(efield_ft)

    def check_efield_resolution(self,pulse_number,*,plot_fields = False):
        """Checks the resolution in time- and frequency-domain of the specified
            pulse, and also optionally plots the fields for view

        Args:
            pulse_number (int) : which pulse to check

        Keyword Args:
            plot_fields (bool) : if True, plots the specified pulse in time- and
                frequency-domains
        """
        efield = self.efields[pulse_number]
        if len(efield) == 1:
            return None
        
        efield_tail = np.max(np.abs([efield[0],efield[-1]]))

        if efield_tail > np.max(np.abs(efield))/100:
            warnings.warn('''Consider using larger time interval, pulse does not decay
                              to less than 1% of maximum value in time domain''')
        efield_ft = self.efield_fts[pulse_number]
        efield_ft_tail = np.max(np.abs([efield_ft[0],efield_ft[-1]]))
        
        if efield_ft_tail > np.max(np.abs(efield_ft))/100:
            warnings.warn('''Consider using smaller value of dt, pulse does not decay 
                              to less than 1% of maximum value in frequency domain''')

        if plot_fields:
            self.plot_efield(pulse_number)

    def plot_efield(self,pulse_number):
        """Plots the specified pulse in both time and frequency domain

        Args:
            pulse_number (int) : which pulse to plot
        """
        efield_t = self.efield_times[pulse_number]
        efield = self.efields[pulse_number]
        efield_w = self.efield_omegas[pulse_number]
        efield_ft = self.efield_fts[pulse_number]
        fig, axes = plt.subplots(1,2)
        l1,l2, = axes[0].plot(efield_t,np.real(efield),efield_t,np.imag(efield))
        plt.legend([l1,l2],['Real','Imag'])
        axes[1].plot(efield_w,np.real(efield_ft),efield_w,np.imag(efield_ft))

        axes[0].set_ylabel('Electric field Amp')
        axes[0].set_xlabel('Time ($\omega_0^{-1})$')
        axes[1].set_xlabel('Frequency ($\omega_0$)')

        fig.suptitle('Check that efield is well-resolved in time and frequency')
        plt.show()

    def set_polarization_sequence(self,polarization_list,*,
                                  reset_calculations=True):
        """Sets the sequences used for either parallel or crossed pump and probe
        
        Args:
            polarization_list (list): list of strings, can be 'x','y', or 
                'z' for linear polarizations or 'r' and 'l' for right and 
                left circularly polarized light, respectively
        Returns:
            None: sets the attribute polarization sequence
"""

        x = np.array([1,0,0])
        y = np.array([0,1,0])
        z = np.array([0,0,1])
        r = np.array([1,-1j,0])/np.sqrt(2)
        l = np.conjugate(r)
        pol_options = {'x':x,'y':y,'z':z,'r':r,'l':l}

        self.polarization_sequence = [pol_options[pol] for pol in polarization_list]
        if reset_calculations:
            self.reset()

    def set_impulsive_pulses(self,phase_discrimination):
        """Automatically sets L impulsive pulses, where L is the length of the 
            input phase discrimination

        Args:
            phase_discrimination (tuple) : tuple of tuples, ((n0r,n0c),(n1r,n1c),...)
                that specify the number of interactions with the rotating and
                counter-rotating terms of each pulse
        """
        L = len(phase_discrimination) # number of pulses
        # Delta = 10 and M = 41 hard-coded in
        efield_t = np.array([0])
        times = [efield_t] * L
        self.set_polarization_sequence(['x'] * L)
        centers = [0] * L
        ef = np.array([1])
        efields = [ef] * L

        self.set_efields(times,efields,centers,phase_discrimination,
                         reset_calculations = True,plot_fields = False)
        

    def set_identical_gaussians(self,sigma_t,c,phase_discrimination,*,
                                Delta = 10, M = 41):
        """Automatically sets L identical gaussian pulses, where L is the length 
            of the input phase discrimination

        Args:
            sigma_t (float) : standard deviation of the Gaussian shape in time-domain
            c (float) : center frequency of the pulse
            phase_discrimination (tuple) : tuple of tuples, ((n0r,n0c),(n1r,n1c),...)
                that specify the number of interactions with the rotating and
                counter-rotating terms of each pulse
        Keyword Args:
            Delta (float) : full time duration for the pulses, in multiples of sigma_t
            M (int) : number of time points to use for the pulses
        """
        L = len(phase_discrimination) # number of pulses
        if self.method == 'chebyshev':
            efield_t = npch.chebpts1(M) * Delta/2 * sigma_t
            dom = np.array([-Delta/2*sigma_t,Delta/2*sigma_t])
            self.doms = [dom] * L
        else:
            efield_t = np.linspace(-Delta/2,Delta/2,num=M)*sigma_t
            
        ef = np.exp(-efield_t**2/(2*sigma_t**2))/(np.sqrt(2*np.pi)*sigma_t)
        times = [efield_t] * L
        self.set_polarization_sequence(['x'] * L)
        centers = [c] * L
        efields = [ef] * L

        self.set_efields(times,efields,centers,phase_discrimination,
                         reset_calculations = True,plot_fields = False)

    def get_closest_index_and_value(self,value,array):
        """Given an array and a desired value, finds the closest actual value
            stored in that array, and returns that value, along with its corresponding 
            array index
        
        Args:
            value (complex) : value to look for in the given array
            array (np.ndarray) : 1D array to look in
"""
        index = np.argmin(np.abs(array - value))
        value = array[index]
        return index, value
    
class UF2BaseClass(BaseClass):
    def save_timing(self):
        save_dict = {'UF2_calculation_time':self.calculation_time}
        np.savez(os.path.join(self.base_path,'UF2_calculation_time.npz'),**save_dict)

class DPBaseClass(BaseClass):
    def save_timing(self):
        save_dict = {'DP_calculation_time':self.calculation_time}
        np.savez(os.path.join(self.base_path,'DP_calculation_time.npz'),**save_dict)
